# (c) 2007-2017 www.scriptol.com 
# Based on PHP FTP Synchronizer by Kim Haskell & Denis Sureau
# Free under the GNU GPL 3 License.

# Scriptol sources are compiled with the Scriptol-JavaScript compiler (www.scriptol.org).
#
# The synchronizer updates a website from a local directory.
# Optionally links are verified by the builtin link checker.

# Read the manual for details of use.

include "path.sol"
include "libftp.js"

bool CHECKMODE = false   // True for virtual operations
bool QUIET = false     // True to display nothing
bool BACKUP = false    // True to work with a backup directory
bool ANYFILES = false  // restoring the site and uploading the full content
bool VERBOSE = false
bool DEBUG = false
bool LINKCHECK = false


text server = "" 	// The ftp address
text user = ""		// login
text pass = ""		// password
text root 
text remotedir
array params = []
text backdir = ""   // backup directory or drive
text temporary = "temporary-file.000.tmp"
text website

int COUNTER = 0     // Number of files uploaded
int FTPCOUNT = 0      // number of file sent
int ERRCOUNT = 0       // upload errors
int falsecounter = 0    // Number of files to copy
int problem = 0

dict OPTIONS = {}
array FilesArray = []


void usage()
	print
	print "JS FTP Synchronizer 1.1 - (c) 2007-2017 Scriptol.com"
	print "----------------------------------------------------"
	print "Syntax:"
	print "  solj sync [options & parameters] source"
	print "Options:"
	print "  -t test mode, display only and do not upload."
	print "  -a upload all files, restore the full site."
	print "  -s speed up, do not check links in pages."
	print "  -v verbose, display more infos."
	print "  -q quiet, display nothing."	
	print "Parameters:"
	print "  -wwebsite url."
	print "  -ppassword."
	print "  -llogin."
	print "  -fftpadr ftp address in the form ftp.domain.tld (like ftp.scriptol.com)."
	print "  -ddirectory remote directory where to upload the files."
	print "  -bbackup, defining a backup directory."  
	print "  source: a directory to backup"
	print "You will be prompted for each required parameter if omitted."
	print "See manual for more."
	exit(0)
return


// Check the presence of the external device or the existence of the directory
// for the backup.

void backError(text b)
  print "Can't write on backup $b, check device or path and try again..."
  exit(0)
return  
  

void checkBackup(text bpath)
    text tempfile = Path.merge(bpath, 'ftpsynxyz.$$$')
    file f 
    f.open(tempfile, "w")
    error ? backError(bpath)
    int saved = f.write("ftp sync check")
    f.close()
    unlink(tempfile)
    if saved = 0
        backError(bpath)
    /if
return


// send a file

void cbFileSent(text err, text data) 
    if err
        print err
        return
    /if    
    if VERBOSE print data
return

// callback for login

void authBack(text err, text res)
    if err 
        print err
        return
    /if
    if VERBOSE print res
return


# Check if file modified

bool backupIdentical(text locfile, text bakfile)
    if DEBUG = true print "Comparing $locfile and local $bakfile"
    if not file_exists(bakfile) return false
	if filesize(locfile) != filesize(bakfile) return false
	if filetime(locfile) =  filetime(bakfile) return true
	array x, y
	x.load(locfile)
	y.load(bakfile)
return x = y


# synchronize

async void synchro(text locdir, text bdir, text hostdir)

	if VERBOSE ? print "Entering $locdir"
	
    array content = scandir(locdir)
    text src, bfile, rmt, name

    if content.empty() return

    // processing files
    
    for name in content

        if name[ .. 10] = "javascript:" continue
        if name[0] = "/"
           src = Path.merge(root,  name) 
        else    
           src = Path.merge(locdir, name)
        /if  

        if filetype(src) = "file"
            rmt = Path.merge(hostdir, name)
  	        // Scan page for checking links
            if ANYFILES = true     // uploading the full content
                bfile = Path.merge(bdir, name)
	            await fileCopy(src, rmt, hostdir, bfile, bdir)
                dict fgroup = { "lfile": src, "rfile": rmt, "hdir": hostdir }
                FilesArray.push(fgroup)
                continue
            /if

           // .htaccess and such special files are ignored and must be changer manually
                     
            if name[0] = "."     
                if not QUIET print name, "skipped"
                problem + 1
                continue 
            /if
			
            // add to list to be compared with backup or remote file and uploaded if different
            if BACKUP = true 
                bfile = Path.merge(bdir, name)
                bool samecont = backupIdentical(src, bfile)
                if  samecont  continue
            /if            

            await fileCopy(src, rmt, hostdir, bfile, bdir)            

            dict fgroup = { "lfile": src, "rfile": rmt, "hdir": hostdir }
            FilesArray.push(fgroup)
	    /if
   /for

    // processing subdirs
	
   for name in content
       if name[0] = '.' continue
       src = Path.merge(locdir, name)	
       if filetype(src) = "dir"
         synchro(src, Path.merge(bdir, name), Path.merge(hostdir, name))
       /if
   /for	

return


# Parsing command line parameters

void processCommand(int argnum, array argv)

    text daystring = ""
	text opt = ""

	if argnum <  2 ? usage()

	for text param in argv

		if param.length() > 1
			opt = param[..1]
		else
			usage()
		/if	
		
		if opt = "-a" 
			ANYFILES = true
			continue
		/if	

		if opt = "-p"
			pass = param[ 2 .. ]
			if pass = nil let die("-p must be followed by the password.")
			continue
		/if	

		if opt = "-l"
			user = param[ 2 .. ]
			if user = nil let die("-l must be followed by the login.")			
			continue
		/if

		if opt = "-f"
			server = param[ 2 .. ]
			if server = nil let die("-f must be followed by the ftp address.")			
			continue
		/if

		if opt = "-d" 
			remotedir = param[ 2 .. ]
			if remotedir = nil let die("-d requires a sub-directory.")
			continue
		/if	

		if opt = "-b"
			backdir = param[ 2 .. ]
			if backdir = nil let die("-b requires a directory.")
			BACKUP = true
			continue
		/if 		
		
		if opt = "-t" 
			CHECKMODE = true
			continue
		/if			

		if opt = "-u" 
			LINKCHECK = true
			continue
		/if			

    	
		if opt = "-w"
			website = param[ 2 .. ]
			if website = nil let die("-w must be followed by the site url.")			
			continue
		/if

		if param[ .. 3] = "ftp."
			server = param
			continue
		/if	
		
		if opt = "-v" 
			VERBOSE = true
			continue
		/if
        
		if opt = "-q" 
			QUIET = true
			continue
		/if	

		if opt = "-~" 
			DEBUG = true
			continue
		/if		
		
		if param[0] = "-" 
            print "Unknown command $param"  
            usage()
        /if   
		
		if root = nil
			root = param
			continue
		/if	
		
		print "Unknown command $param"
    
    usage()
		
	/for


    if BACKUP = true let checkBackup(backdir)

	if server = nil input "FTP location: ",  server
	if server = nil let exit(0)

	if root = nil input "Directory to send: ",  root
	if root = nil let exit(0)

	if user = nil input "Login: ",  user
	if user = nil let exit(0)

	if pass = nil input "Password: ", pass	
	if pass = nil let exit(0)
	
return


void saybye(text err, text data) 
    if err
        print err
        return
    /if    
    print "Done!"
return

void dispCount()
	echo COUNTER, " file", plural(COUNTER), " copied"
	if CHECKMODE
        if falsecounter > 0 
            echo ", ", falsecounter, " file", plural(falsecounter)," to update"
        else
            echo ", nothing to update"
        /if
    /if   
    print "."
return


int main(int argc, array argv)

	array x = argv[ 1 .. ]
	processCommand(argc, x)
    problem = 0

    if not QUIET
        if VERBOSE = true print "Verbose mode enabled"
        if DEBUG = true print "Debug mode enabled"
        print "Source directory: $root"
        print "Remote directory:", remotedir
        if BACKUP = true print "Backup location: $backdir"
        if ANYFILES = true print "Website will be restored"
        if LINKCHECK = true print "Link checker not implemented yet"
    /if

    OPTIONS = { 
            "host": server, 
            "port": 21, 
            "user": user, 
            "pass": pass,
            "debugMode": true
    }

	print "Synchronizing $root on $server"
	
	synchro(root, backdir, remotedir)	
    beforeExit(dispCount)

return 0

main($argc, $argv)
